print("import: 'seabreeze'")
import seabreeze

print("import: 'seabreeze.types'")
import seabreeze.types

print("import: 'seabreeze.cseabreeze'")
import seabreeze.cseabreeze

print("import: 'seabreeze.pyseabreeze'")
import seabreeze.pyseabreeze

